﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Text;
using System.Text.RegularExpressions;
using System.Web;
using System.Web.Mvc;

namespace MvcExtensions {

    /// <summary>
    /// Extends the HtmlHelper class to add functionality to insert content
    /// in different areas, regardless of scope
    /// </summary>
    public static class ContentMarkerHtmlHelper {

        #region Helper Methods

        /// <summary>
        /// Inserts a new marker using the identity provided
        /// </summary>
        public static void InsertMarker(this HtmlHelper helper, string identity) {
            _StreamController.Filter.Insert(identity);
        }

        /// <summary>
        /// Appends content to the specified marker. If the marker does not exist, the
        /// content will be saved until the marker is added
        /// </summary>
        public static void AppendToMarker(this HtmlHelper helper, string identity, string content) {
            _StreamController.Filter.AppendTo(identity, content);
        }

        /// <summary>
        /// Inserts a new marker using the identity of the enumerated type
        /// </summary>
        public static void InsertMarker(this HtmlHelper helper, Enum identity) {
            _StreamController.Filter.Insert(Enum.GetName(identity.GetType(), identity));
        }

        /// <summary>
        /// Appends content to the specified marker. If the marker does not exist, the
        /// content will be saved until the marker is added
        /// </summary>
        public static void AppendToMarker(this HtmlHelper helper, Enum identity, string content) {
            _StreamController.Filter.AppendTo(
                Enum.GetName(identity.GetType(), identity),
                content
                );
        }

        #endregion

        #region Stream Controller Class

        //handles catching and formatting data before it is
        //send out as rendered HTML
        private class _StreamController : MemoryStream {

            #region Constants

            private const string REGEX_MATCH_MARKERS = @"\{{3}\:{2}(?<name>\w+)\:{2}\}{3}";
            private const string REGEX_MATCH_GROUP_NAME = "name";
            private const string MARKER_START = "{{{::";
            private const string MARKER_END = "::}}}";

            #endregion

            #region Constructors

            public _StreamController() {
                this._Output = new StringBuilder();
                this._Markers = new Dictionary<string, string>();
            }

            #endregion

            #region Static Methods

            //verifies the filter has been setup to catch the markers
            public static _StreamController Filter {
                get {
                    if (!(HttpContext.Current.Response.Filter is _StreamController)) {
                        HttpContext.Current.Response.Filter = new _StreamController();
                    }
                    return (_StreamController)HttpContext.Current.Response.Filter;
                }
            }

            #endregion

            #region Members

            StringBuilder _Output;
            Dictionary<string, string> _Markers;

            #endregion

            #region Methods

            //override the write event to capture the string
            public override void Write(byte[] buffer, int offset, int count) {
                string content = HttpContext.Current.Response.ContentEncoding.GetString(buffer, offset, count);
                _Output.Append(content);
            }

            //inserts a marker into the content to replace later
            public void Insert(string identity) {
                string marker = string.Concat(
                    _StreamController.MARKER_START,
                    identity,
                    _StreamController.MARKER_END
                    );
                HttpContext.Current.Response.Write(marker);
            }

            //append content that will be used to replace a marker
            public void AppendTo(string identity, string content) {
                if (!_Markers.ContainsKey(identity)) {
                    _Markers.Add(identity, string.Empty);
                }
                _Markers[identity] = string.Concat(_Markers[identity], content);
            }

            #endregion

            #region Overrides

            //overrides the dispose event to render our final content
            protected override void Dispose(bool disposing) {

                //if this is disposing, replace the content now
                if (disposing) {
                    string render = _Output.ToString();

                    //loop backwards and replace the markers
                    MatchCollection matches = Regex.Matches(render, _StreamController.REGEX_MATCH_MARKERS);
                    for (int i = matches.Count; i-- > 0; ) {
                        Match match = matches[i];

                        //find the replacement
                        string key = match.Groups[_StreamController.REGEX_MATCH_GROUP_NAME].Value;
                        string replacement = string.Empty;
                        if (this._Markers.ContainsKey(key)) {
                            replacement = this._Markers[key];
                        }

                        //remove the match and insert the content
                        render = render.Remove(match.Index, match.Length);
                        render = render.Insert(match.Index, replacement);
                    }

                    //write the final content to the output
                    HttpContext.Current.Response.Write(render);
                }
            }

            #endregion

        }

        #endregion

    }

    /// <summary>
    /// Enumerated sections for the document
    /// </summary>
    public enum Document {
        Head,
        Body,
        Script,
        Style,
        Title,
        Content,
        Footer,
        Menu,
        Navigation
    }

}
